import json
import logging
from datetime import datetime, timedelta
from typing import Union, Mapping, NewType, Type, Dict, Callable, Any, Tuple, List

logger = logging.getLogger("pykusto")

KustoTypes = Union[str, int, bool, datetime, Mapping, List, Tuple, float, timedelta]
# TODO: Unhandled date types: guid, decimal

KQL = NewType('KQL', str)


def datetime_to_kql(dt: datetime) -> KQL:
    return KQL(dt.strftime('datetime(%Y-%m-%d %H:%M:%S.%f)'))


def timedelta_to_kql(td: timedelta) -> KQL:
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return KQL('time({days}.{hours}:{minutes}:{seconds}.{microseconds})'.format(
        days=td.days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=td.microseconds,
    ))


def dynamic_to_kql(d: Union[Mapping, List, Tuple]) -> KQL:
    query = list(json.dumps(d))
    # Issue #11
    counter = 0
    prev = ""
    for i, c in enumerate(query):
        if counter == 0:
            if c == "[":
                query[i] = "("
            elif c == "]":
                query[i] = ")"
            elif c in ['"', '\''] and prev != "\\":
                counter += 1
        elif counter > 0:
            if c in ['"', '\''] and prev != "\\":
                counter -= 1
        prev = query[i]
    assert counter == 0
    return KQL("".join(query))


def bool_to_kql(b: bool) -> KQL:
    return KQL('true') if b else KQL('false')


def str_to_kql(s: str) -> KQL:
    return KQL('"{}"'.format(s))


KQL_CONVERTER_BY_TYPE: Dict[Type, Callable[[Any], KQL]] = {
    datetime: datetime_to_kql,
    timedelta: timedelta_to_kql,
    Mapping: dynamic_to_kql,
    List: dynamic_to_kql,
    Tuple: dynamic_to_kql,
    bool: bool_to_kql,
    str: str_to_kql,
    int: KQL,
    float: KQL,
}


def to_kql(obj: KustoTypes) -> KQL:
    for kusto_type, converter in KQL_CONVERTER_BY_TYPE.items():
        if isinstance(obj, kusto_type):
            return converter(obj)
    raise ValueError("No KQL converter found for object {} of type {}".format(obj, type(obj)))



from pykusto import utils
from test.test_base import TestBase


class TestUtils(TestBase):
    def test_dynamic_to_kql(self):
        dict ={
            "name": "Alan",
            "age": 21,
            "address": ("NY", 36),
            "pets": ["Libby", "Panda", "]", "["]
        }
        self.assertEqual(
            "{\"name\": \"Alan\", \"age\": 21, \"address\": (\"NY\", 36), \"pets\": (\"Libby\", \"Panda\", \"]\", \"[\")}",
            utils.to_kql(dict)
        )



from datetime import timedelta

from pykusto.expressions import column_generator as col
from pykusto.query import Query
from test.test_base import TestBase


class TestExpressions(TestBase):
    def test_contains(self):
        self.assertEqual(
            ' | where foo contains "bar"',
            Query().where(col.foo.contains('bar')).render(),
        )

    def test_array_access(self):
        self.assertEqual(
            ' | where (arr[3]) == "bar"',
            Query().where(col.arr[3] == 'bar').render(),
        )

    def test_array_access_expression_index(self):
        self.assertEqual(
            ' | where (arr[(foo * 2)]) == "bar"',
            Query().where(col.arr[col.foo * 2] == 'bar').render(),
        )

    def test_mapping_access(self):
        self.assertEqual(
            ' | where (dict["key"]) == "bar"',
            Query().where(col.dict['key'] == 'bar').render(),
        )

    def test_mapping_access_expression_index(self):
        self.assertEqual(
            ' | where (dict[foo]) == "bar"',
            Query().where(col.dict[col.foo] == 'bar').render(),
        )

    def test_assign_to(self):
        self.assertEqual(
            " | extend ['foo.bar'] = (shoo * 2)",
            Query().extend((col.shoo * 2).assign_to(col.foo.bar)).render(),
        )

    def test_between_timespan(self):
        self.assertEqual(
            " | where foo between (time(0.0:0:0.0) .. time(0.3:0:0.0))",
            Query().where(col.foo.between(timedelta(0), timedelta(hours=3))).render(),
        )

    def test_is_empty(self):
        self.assertEqual(
            'isempty(foo)',
            col.foo.is_empty().kql,
        )

    def test_method_does_not_exist(self):
        self.assertRaises(
            AttributeError,
            col.foo.non_existant_method,
        )

    def test_column_generator(self):
        self.assertEqual(
            " | project ['foo.bar']",
            Query().project(col.foo.bar).render(),
        )

    def test_column_generator_2(self):
        self.assertEqual(
            " | project ['foo.bar']",
            Query().project(col['foo.bar']).render(),
        )

    def test_is_in(self):
        self.assertEqual(
            " | where foo in (\"A\", \"B\", \"C\")",
            Query().where(col.foo.is_in(["A", "B", "C"])).render()
        )
        self.assertEqual(
            " | where foo in (\"[\", \"[[\", \"]\")",
            Query().where(col.foo.is_in(['[', "[[", "]"])).render()
        )


import logging
import sys
from unittest import TestCase

from pykusto.utils import logger


class TestBase(TestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)5s %(message)s'
        )

    def setUp(self):
        logger.info("Running test: " + self._testMethodName)
